#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Dictionaries and classes defined under the W3C WoT Scripting API specification.

.. autosummary::
    :toctree: _wot

    wotpy.wot.consumed
    wotpy.wot.dictionaries
    wotpy.wot.discovery
    wotpy.wot.exposed
    wotpy.wot.constants
    wotpy.wot.enums
    wotpy.wot.events
    wotpy.wot.form
    wotpy.wot.interaction
    wotpy.wot.servient
    wotpy.wot.td
    wotpy.wot.thing
    wotpy.wot.validation
    wotpy.wot.wot
"""
